@@ -1,7 +1,7 @@
 /* cgmanager-proxy
  *
- * Copyright © 2013 Stphane Graber
- * Author: Stphane Graber <stgraber@ubuntu.com>
+ * Copyright © 2013 Stephane Graber
+ * Author: Stephane Graber <stgraber@ubuntu.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2, as
@@ -45,6 +45,22 @@ bool master_running(void)
 	return false;
 }
 
+/* wait up to 2 seconds for a reply from cgmanager */
+static int proxyrecv(int sockfd, void *buf, size_t len)
+{
+	struct timeval tv;
+	fd_set rfds;
+
+	FD_ZERO(&rfds);
+	FD_SET(sockfd, &rfds);
+	tv.tv_sec = 2;
+	tv.tv_usec = 0;
+
+	if (select(sockfd+1, &rfds, NULL, NULL, &tv) < 0)
+		return -1;
+	return recv(sockfd, buf, len, MSG_DONTWAIT);
+}
+
 static void cgm_dbus_disconnected(DBusConnection *connection);
 
 int setup_proxy(void)
@@ -114,6 +130,8 @@ int setup_proxy(void)
 		nih_fatal("Failed to open connection to %s: %s",
 				CGPROXY_DBUS_PATH, err->message);
 		nih_free(err);
+
+		return -1;
 	}
 	return 0;
 }
@@ -143,6 +161,7 @@ bool send_dummy_msg(DBusConnection *conn)
 	DBusMessage *message = NULL;
 	DBusMessageIter iter;
 	int a;
+
 	message = dbus_message_new_method_call(dbus_bus_get_unique_name(conn),
 			"/org/linuxcontainers/cgmanager",
 			"org.linuxcontainers.cgmanager0_0", "Ping");
@@ -207,7 +226,7 @@ static bool complete_dbus_request(DBusMessage *message,
 	dbus_connection_flush(server_conn);
 	dbus_message_unref(message);
 
-	if (recv(sv[0], buf, 1, 0) != 1) {
+	if (proxyrecv(sv[0], buf, 1) != 1) {
 		nih_error("%s: Error getting reply from server over socketpair",
 			  __func__);
 		return false;
@@ -221,7 +240,7 @@ static bool complete_dbus_request(DBusMessage *message,
 	if (!vcred) // this request only requires one scm_credential
 		return true;
 
-	if (recv(sv[0], buf, 1, 0) != 1) {
+	if (proxyrecv(sv[0], buf, 1) != 1) {
 		nih_error("%s: Error getting reply from server over socketpair",
 			__func__);
 		return false;
@@ -272,8 +291,57 @@ int get_pid_cgroup_main (void *parent, const char *controller,
 		goto out;
 	}
 
-	// TODO - switch to nih_io_message_recv?
-	if (recv(sv[0], s, MAXPATHLEN-1, 0) <= 0)
+	if (proxyrecv(sv[0], s, MAXPATHLEN-1) <= 0)
+		nih_error("%s: Error reading result from cgmanager",
+			__func__);
+	else {
+		*output = NIH_MUST( nih_strdup(parent, s) );
+		ret = 0;
+	}
+out:
+	close(sv[0]);
+	close(sv[1]);
+	return ret;
+}
+
+int get_pid_cgroup_abs_main (void *parent, const char *controller,
+		struct ucred p, struct ucred r, struct ucred v, char **output)
+{
+	DBusMessage *message;
+	DBusMessageIter iter;
+	int sv[2], ret = -1;
+	char s[MAXPATHLEN] = { 0 };
+
+	if (memcmp(&p, &r, sizeof(struct ucred)) != 0) {
+		nih_error("%s: proxy != requestor", __func__);
+		return -1;
+	}
+
+	if (!(message = start_dbus_request("GetPidCgroupAbsScm", sv))) {
+		nih_error("%s: error starting dbus request", __func__);
+		return -1;
+	}
+
+	dbus_message_iter_init_append(message, &iter);
+	if (!dbus_message_iter_append_basic (&iter,
+			DBUS_TYPE_STRING,
+			&controller)) {
+		dbus_message_unref(message);
+		nih_error("%s: out of memory", __func__);
+		goto out;
+	}
+	if (! dbus_message_iter_append_basic (&iter, DBUS_TYPE_UNIX_FD, &sv[1])) {
+		dbus_message_unref(message);
+		nih_error("%s: out of memory", __func__);
+		goto out;
+	}
+
+	if (!complete_dbus_request(message, sv, &r, &v)) {
+		nih_error("%s: error completing dbus request", __func__);
+		goto out;
+	}
+
+	if (proxyrecv(sv[0], s, MAXPATHLEN-1) <= 0)
 		nih_error("%s: Error reading result from cgmanager",
 			__func__);
 	else {
@@ -333,7 +401,7 @@ int do_move_pid_main (const char *controller, const char *cgroup,
 		goto out;
 	}
 
-	if (recv(sv[0], buf, 1, 0) == 1 && *buf == '1')
+	if (proxyrecv(sv[0], buf, 1) == 1 && *buf == '1')
 		ret = 0;
 out:
 	close(sv[0]);
@@ -359,10 +427,21 @@ int move_pid_main (const char *controller, const char *cgroup,
 int move_pid_abs_main (const char *controller, const char *cgroup,
 		struct ucred p, struct ucred r, struct ucred v)
 {
+#if 0
+	/*
+	 * We used to enforce that r must be root.  However that's
+	 * overly restrictive.
+	 * Cgmanager ensures that r must have write access to the
+	 * tasks file.  That seems sufficient.  However if it is deemed
+	 * insufficient, we can ensure that r's user or group id own
+	 * all parent directories up to a common parent, from v.cgroup
+	 * to the requested cgroup.  THIS CODE does NOT do that.
+	 */
 	if (r.uid) {
 		nih_error("%s: uid %u tried to escape", __func__, r.uid);
 		return -1;
 	}
+#endif
 	if (!sane_cgroup(cgroup)) {
 		nih_error("%s: unsafe cgroup", __func__);
 		return -1;
@@ -415,7 +494,7 @@ int create_main (const char *controller, const char *cgroup, struct ucred p,
 		goto out;
 	}
 
-	if (recv(sv[0], buf, 1, 0) == 1 && (*buf == '1' || *buf == '2'))
+	if (proxyrecv(sv[0], buf, 1) == 1 && (*buf == '1' || *buf == '2'))
 		ret = 0;
 	*existed = *buf == '2' ? 1 : -1;
 out:
@@ -469,7 +548,7 @@ int chown_main (const char *controller, const char *cgroup,
 		goto out;
 	}
 
-	if (recv(sv[0], buf, 1, 0) == 1 && *buf == '1')
+	if (proxyrecv(sv[0], buf, 1) == 1 && *buf == '1')
 		ret = 0;
 out:
 	close(sv[0]);
@@ -532,7 +611,7 @@ int chmod_main (const char *controller, const char *cgroup, const char *file,
 		goto out;
 	}
 
-	if (recv(sv[0], buf, 1, 0) == 1 && *buf == '1')
+	if (proxyrecv(sv[0], buf, 1) == 1 && *buf == '1')
 		ret = 0;
 out:
 	close(sv[0]);
@@ -590,7 +669,7 @@ int get_value_main (void *parent, const char *controller, const char *cgroup,
 		goto out;
 	}
 
-	if (recv(sv[0], output, MAXPATHLEN, 0) <= 0) {
+	if (proxyrecv(sv[0], output, MAXPATHLEN) <= 0) {
 		nih_error("%s: Failed reading string from cgmanager: %s",
 			__func__, strerror(errno));
 	} else {
@@ -659,7 +738,7 @@ int set_value_main (const char *controller, const char *cgroup,
 		goto out;
 	}
 
-	if (recv(sv[0], buf, 1, 0) == 1 && *buf == '1')
+	if (proxyrecv(sv[0], buf, 1) == 1 && *buf == '1')
 		ret = 0;
 out:
 	close(sv[0]);
@@ -717,7 +796,7 @@ int remove_main (const char *controller, const char *cgroup, struct ucred p,
 		goto out;
 	}
 
-	if (recv(sv[0], buf, 1, 0) == 1 && (*buf == '1' || *buf == '2'))
+	if (proxyrecv(sv[0], buf, 1) == 1 && (*buf == '1' || *buf == '2'))
 		ret = 0;
 	*existed = *buf == '2' ? 1 : -1;
 out:
@@ -772,7 +851,83 @@ int get_tasks_main (void *parent, const char *controller, const char *cgroup,
 		nih_error("%s: error completing dbus request", __func__);
 		goto out;
 	}
-	if (recv(sv[0], &nrpids, sizeof(uint32_t), 0) != sizeof(uint32_t))
+	if (proxyrecv(sv[0], &nrpids, sizeof(uint32_t)) != sizeof(uint32_t))
+		goto out;
+	if (nrpids == -1) {
+		nih_error("%s: bad cgroup: %s:%s", __func__, controller, cgroup);
+		ret = -1;
+		goto out;
+	}
+	if (nrpids == 0) {
+		ret = 0;
+		goto out;
+	}
+
+	*pids = NIH_MUST( nih_alloc(parent, nrpids * sizeof(uint32_t)) );
+	for (i=0; i<nrpids; i++) {
+		get_scm_creds_sync(sv[0], &tcred);
+		if (tcred.pid == -1) {
+			nih_error("%s: Failed getting pid from server",
+				__func__);
+			goto out;
+		}
+		(*pids)[i] = tcred.pid;
+	}
+	ret = nrpids;
+out:
+	close(sv[0]);
+	close(sv[1]);
+	return ret;
+}
+
+int get_tasks_recursive_main (void *parent, const char *controller,
+		const char *cgroup, struct ucred p, struct ucred r,
+		int32_t **pids)
+{
+	DBusMessage *message;
+	DBusMessageIter iter;
+	int sv[2], ret = -1;
+	uint32_t nrpids;
+	struct ucred tcred;
+	int i;
+
+	if (memcmp(&p, &r, sizeof(struct ucred)) != 0) {
+		nih_error("%s: proxy != requestor", __func__);
+		return -1;
+	}
+
+	if (!sane_cgroup(cgroup)) {
+		nih_error("%s: unsafe cgroup", __func__);
+		return -1;
+	}
+
+	if (!(message = start_dbus_request("GetTasksRecursiveScm", sv))) {
+		nih_error("%s: error starting dbus request", __func__);
+		return -1;
+	}
+
+	dbus_message_iter_init_append(message, &iter);
+	if (! dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &controller)) {
+		nih_error("%s: out of memory", __func__);
+		dbus_message_unref(message);
+		goto out;
+	}
+	if (! dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &cgroup)) {
+		nih_error("%s: out of memory", __func__);
+		dbus_message_unref(message);
+		goto out;
+	}
+	if (! dbus_message_iter_append_basic (&iter, DBUS_TYPE_UNIX_FD, &sv[1])) {
+		nih_error("%s: out of memory", __func__);
+		dbus_message_unref(message);
+		goto out;
+	}
+
+	if (!complete_dbus_request(message, sv, &r, NULL)) {
+		nih_error("%s: error completing dbus request", __func__);
+		goto out;
+	}
+	if (proxyrecv(sv[0], &nrpids, sizeof(uint32_t)) != sizeof(uint32_t))
 		goto out;
 	if (nrpids == -1) {
 		nih_error("%s: bad cgroup: %s:%s", __func__, controller, cgroup);
@@ -851,7 +1006,7 @@ int list_children_main (void *parent, const char *controller, const char *cgroup
 		goto out;
 	}
 
-	if (recv(sv[0], &nrkids, sizeof(int32_t), 0) != sizeof(int32_t))
+	if (proxyrecv(sv[0], &nrkids, sizeof(int32_t)) != sizeof(int32_t))
 		goto out;
 	if (nrkids == 0) {
 		ret = 0;
@@ -862,20 +1017,27 @@ int list_children_main (void *parent, const char *controller, const char *cgroup
 		ret = -1;
 		goto out;
 	}
-	if (recv(sv[0], &len, sizeof(uint32_t), 0) != sizeof(uint32_t))
+	if (proxyrecv(sv[0], &len, sizeof(uint32_t)) != sizeof(uint32_t))
 		goto out;
 
-	paths = nih_alloc(NULL, len);
+	paths = nih_alloc(NULL, len+1);
+	paths[len] = '\0';
 	if (read(sv[0], paths, len) != len) {
 		nih_error("%s: Failed getting paths from server", __func__);
 		goto out;
 	}
 
 	*output = NIH_MUST( nih_alloc(parent, sizeof( char*)*(nrkids+1)) );
+	memset(*output, 0, (nrkids + 1) * sizeof(char *));
 
 	s = paths;
-	(*output)[nrkids] = NULL;
 	for (i=0; i<nrkids; i++) {
+		if (s > paths + len) {
+			ret = -1;
+			nih_error("%s: corrupted result from cgmanager",
+					__func__);
+			goto out;
+		}
 		(*output)[i] = NIH_MUST( nih_strdup(parent, s) );
 		s += strlen(s) + 1;
 	}
@@ -931,7 +1093,7 @@ int remove_on_empty_main (const char *controller, const char *cgroup,
 		goto out;
 	}
 
-	if (recv(sv[0], buf, 1, 0) == 1 && (*buf == '1'))
+	if (proxyrecv(sv[0], buf, 1) == 1 && (*buf == '1'))
 		ret = 0;
 out:
 	close(sv[0]);
@@ -939,6 +1101,260 @@ out:
 	return ret;
 }
 
+int prune_main (const char *controller, const char *cgroup,
+		struct ucred p, struct ucred r)
+{
+	DBusMessage *message;
+	DBusMessageIter iter;
+	int sv[2], ret = -1;
+	char buf[1];
+
+	if (memcmp(&p, &r, sizeof(struct ucred)) != 0) {
+		nih_error("%s: proxy != requestor", __func__);
+		return -1;
+	}
+
+	if (!sane_cgroup(cgroup)) {
+		nih_error("%s: unsafe cgroup", __func__);
+		return -1;
+	}
+
+	if (!(message = start_dbus_request("PruneScm", sv))) {
+		nih_error("%s: error starting dbus request", __func__);
+		return -1;
+	}
+
+	dbus_message_iter_init_append(message, &iter);
+	if (! dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &controller)) {
+		nih_error("%s: out of memory", __func__);
+		dbus_message_unref(message);
+		goto out;
+	}
+	if (! dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &cgroup)) {
+		nih_error("%s: out of memory", __func__);
+		dbus_message_unref(message);
+		goto out;
+	}
+	if (! dbus_message_iter_append_basic (&iter, DBUS_TYPE_UNIX_FD, &sv[1])) {
+		nih_error("%s: out of memory", __func__);
+		dbus_message_unref(message);
+		goto out;
+	}
+
+	if (!complete_dbus_request(message, sv, &r, NULL)) {
+		nih_error("%s: error completing dbus request", __func__);
+		goto out;
+	}
+
+	if (proxyrecv(sv[0], buf, 1) == 1 && (*buf == '1'))
+		ret = 0;
+out:
+	close(sv[0]);
+	close(sv[1]);
+	return ret;
+}
+
+int list_controllers_main (void *parent, char ***output)
+{
+	DBusMessage *message = NULL, *reply = NULL;
+	char **         output_local = NULL;
+	DBusError       error;
+	DBusMessageIter iter;
+	int		ret = -1;
+	DBusMessageIter output_local_iter;
+	size_t          output_local_size;
+
+	*output = NULL;
+	message = dbus_message_new_method_call(dbus_bus_get_unique_name(server_conn),
+			"/org/linuxcontainers/cgmanager",
+			"org.linuxcontainers.cgmanager0_0", "ListControllers");
+
+	dbus_error_init (&error);
+
+	reply = dbus_connection_send_with_reply_and_block (server_conn, message, -1, &error);
+	if (! reply) {
+		dbus_message_unref (message);
+
+		nih_error("%s: error completing dbus request: %s %s", __func__,
+			error.name, error.message);
+
+		dbus_error_free (&error);
+		return -1;
+	}
+	dbus_message_unref (message);
+
+	dbus_message_iter_init (reply, &iter);
+
+	if (dbus_message_iter_get_arg_type (&iter) != DBUS_TYPE_ARRAY)
+		goto out;
+
+	dbus_message_iter_recurse (&iter, &output_local_iter);
+
+	output_local_size = 0;
+	output_local = NULL;
+
+	output_local = NIH_MUST( nih_alloc (parent, sizeof (char *)) );
+
+	output_local[output_local_size] = NULL;
+
+	while (dbus_message_iter_get_arg_type (&output_local_iter) != DBUS_TYPE_INVALID) {
+		const char *output_local_element_dbus;
+		char **     output_local_tmp;
+		char *      output_local_element;
+
+		if (dbus_message_iter_get_arg_type (&output_local_iter) != DBUS_TYPE_STRING)
+			goto out;
+
+		dbus_message_iter_get_basic (&output_local_iter, &output_local_element_dbus);
+
+		output_local_element = NIH_MUST( nih_strdup (output_local, output_local_element_dbus) );
+
+		dbus_message_iter_next (&output_local_iter);
+
+		output_local_tmp = NIH_MUST( nih_realloc (output_local, parent, sizeof (char *) * (output_local_size + 2)) );
+
+		output_local = output_local_tmp;
+		output_local[output_local_size] = output_local_element;
+		output_local[output_local_size + 1] = NULL;
+
+		output_local_size++;
+	}
+
+	dbus_message_iter_next (&iter);
+
+	if (dbus_message_iter_get_arg_type (&iter) != DBUS_TYPE_INVALID)
+		goto out;
+
+	*output = output_local;
+
+	ret = 0;
+
+out:
+	if (reply)
+		dbus_message_unref (reply);
+	if (ret)
+		nih_free (output_local);
+	return ret;
+}
+
+static char *find_eol(char *s)
+{
+	while (*s && *s != '\n')
+		s++;
+	return s;
+}
+
+int list_keys_main (void *parent, const char *controller, const char *cgroup,
+		    struct ucred p, struct ucred r,
+		    struct keys_return_type ***output)
+{
+	DBusMessage *message;
+	DBusMessageIter iter;
+	int sv[2], ret = -1;
+	uint32_t len;
+	int32_t nrkeys;
+	nih_local char * results = NULL;
+	char *s;
+	int i;
+
+	*output = NULL;
+	if (memcmp(&p, &r, sizeof(struct ucred)) != 0) {
+		nih_error("%s: proxy != requestor", __func__);
+		return -1;
+	}
+
+	if (!sane_cgroup(cgroup)) {
+		nih_error("%s: unsafe cgroup", __func__);
+		return -1;
+	}
+
+	if (!(message = start_dbus_request("ListKeysScm", sv))) {
+		nih_error("%s: error starting dbus request", __func__);
+		return -1;
+	}
+
+	dbus_message_iter_init_append(message, &iter);
+	if (! dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &controller)) {
+		nih_error("%s: out of memory", __func__);
+		dbus_message_unref(message);
+		goto out;
+	}
+	if (! dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &cgroup)) {
+		nih_error("%s: out of memory", __func__);
+		dbus_message_unref(message);
+		goto out;
+	}
+	if (! dbus_message_iter_append_basic (&iter, DBUS_TYPE_UNIX_FD, &sv[1])) {
+		nih_error("%s: out of memory", __func__);
+		dbus_message_unref(message);
+		goto out;
+	}
+
+	if (!complete_dbus_request(message, sv, &r, NULL)) {
+		nih_error("%s: error completing dbus request", __func__);
+		goto out;
+	}
+
+	if (proxyrecv(sv[0], &nrkeys, sizeof(int32_t)) != sizeof(int32_t))
+		goto out;
+	if (nrkeys == 0) {
+		ret = 0;
+		goto out;
+	}
+	if (nrkeys < 0) {
+		nih_error("%s: Server encountered an error: bad cgroup?", __func__);
+		ret = -1;
+		goto out;
+	}
+	if (proxyrecv(sv[0], &len, sizeof(uint32_t)) != sizeof(uint32_t))
+		goto out;
+
+	results = nih_alloc(NULL, len+1);
+	results[len] = '\0';
+	if (read(sv[0], results, len) != len) {
+		nih_error("%s: Failed getting results from server", __func__);
+		goto out;
+	}
+
+	*output = NIH_MUST( nih_alloc(parent, sizeof(**output)*(nrkeys+1)) );
+	memset(*output, 0, (nrkeys + 1) * sizeof(**output));
+
+	s = results;
+	for (i=0; i<nrkeys; i++) {
+		struct keys_return_type *tmp;
+		char *s2 = find_eol(s);
+		if (s2 > results + len)
+			goto bad;
+		*s2 = '\0';
+		(*output)[i] = tmp = NIH_MUST( nih_new(*output, struct keys_return_type) );
+		tmp->name = NIH_MUST( nih_strdup(tmp, s) );
+		s = s2 + 1;
+		s2 = find_eol(s);
+		if (s2 > results + len)
+			goto bad;
+		if (sscanf(s, "%u\n", &tmp->uid) != 1)
+			goto bad;
+		s = s2 + 1;
+		s2 = find_eol(s);
+		if (sscanf(s, "%u\n", &tmp->gid) != 1)
+			goto bad;
+		s = s2 + 1;
+		s2 = find_eol(s);
+		if (sscanf(s, "%u\n", &tmp->perms) != 1)
+			goto bad;
+		s = s2 + 1;
+	}
+	ret = nrkeys;
+out:
+	close(sv[0]);
+	close(sv[1]);
+	return ret;
+
+bad:
+	ret = -1;
+	nih_error("%s: corrupted result from cgmanager", __func__);
+	goto out;
+}
 
 /**
  * options:
@@ -948,6 +1364,8 @@ out:
 static NihOption options[] = {
 	{ 0, "daemon", N_("Detach and run in the background"),
 	  NULL, NULL, &daemonise, NULL },
+	{ 0, "sigstop", N_("Raise SIGSTOP when ready"),
+		NULL, NULL, &sigstop, NULL },
 	{ 0, "check-master", N_("Check whether cgmanager is running"),
 	  NULL, NULL, &checkmaster, NULL },
 
@@ -986,6 +1404,7 @@ main (int argc, char *argv[])
 			exit(0);
 		exit(1);
 	}
+
 	if (setup_proxy() < 0) {
 		nih_fatal ("Failed to set up as proxy");
 		exit(1);
@@ -1030,6 +1449,9 @@ main (int argc, char *argv[])
 		exit(1);
 	}
 
+	if (sigstop)
+		raise(SIGSTOP);
+
 	ret = nih_main_loop ();
 
 	/* Destroy any PID file we may have created */
