@@ -19,6 +19,9 @@
 
 #include <frontend.h>
 #include <sys/resource.h>
+#include <sys/statvfs.h>
+#include <sys/vfs.h>
+#include <linux/fs.h>
 
 /*
  * Maximum depth of directories we allow in Create
@@ -335,7 +338,7 @@ int create_main(const char *controller, const char *cgroup, struct ucred p,
 	char *tok;
 	int ret;
 
-	*existed = 1;
+	*existed = -1;
 	if (!cgroup || ! *cgroup)  // nothing to do
 		return 0;
 
@@ -363,8 +366,8 @@ int create_main(const char *controller, const char *cgroup, struct ucred p,
 			goto next;
 		if (ret != 0)
 			return -1;
-		if (!e)
-			*existed = 0;
+		if (e == 1)
+			*existed = 1;
 next:
 		tok = strtok(NULL, ",");
 	}
@@ -856,16 +859,16 @@ int get_tasks_main(void *parent, const char *controller, const char *cgroup,
 	return nrpids;
 }
 
-int do_collect_tasks(void *parent, char *path, int32_t **pids,
+int do_collect_tasks(void *parent, char **path, int32_t **pids,
 			int *alloced_pids, int *nrpids)
 {
 	struct dirent dirent, *direntp;
 	DIR *dir;
 	const char *key = "tasks";
 
-	dir = opendir(path);
+	dir = opendir(*path);
 	if (!dir) {
-		nih_warn("%s: Failed to open dir %s for recursive deletion", __func__, path);
+		nih_warn("%s: Failed to open dir %s for recursive deletion", __func__, *path);
 		return -2;
 	}
 
@@ -879,12 +882,12 @@ int do_collect_tasks(void *parent, char *path, int32_t **pids,
 		if (!strcmp(direntp->d_name, ".") ||
 		    !strcmp(direntp->d_name, ".."))
 			continue;
-		childname = NIH_MUST( nih_sprintf(NULL, "%s/%s", path, direntp->d_name) );
+		childname = NIH_MUST( nih_sprintf(NULL, "%s/%s", *path, direntp->d_name) );
 		rc = lstat(childname, &mystat);
 		if (rc)
 			continue;
 		if (S_ISDIR(mystat.st_mode))
-			if (do_collect_tasks(parent, childname, pids,
+			if (do_collect_tasks(parent, &childname, pids,
 						alloced_pids, nrpids) == -1)
 				nih_info("%s: error descending subdirs", __func__);
 	}
@@ -892,17 +895,9 @@ int do_collect_tasks(void *parent, char *path, int32_t **pids,
 	closedir(dir);
 
 	/* Get tasks for this directory */
+	NIH_MUST( nih_strcat_sprintf(path, NULL, "/%s", key) );
 
-	/* append the filename */
-	if (strlen(path) + strlen(key) + 2 > MAXPATHLEN) {
-		nih_error("%s: filename too long for cgroup %s key %s", __func__, path, key);
-		return -1;
-	}
-
-	strncat(path, "/", MAXPATHLEN-1);
-	strncat(path, key, MAXPATHLEN-1);
-
-	return file_read_pids(parent, path, pids, alloced_pids, nrpids);
+	return file_read_pids(parent, *path, pids, alloced_pids, nrpids);
 }
 
 int collect_tasks(void *parent, const char *controller, const char *cgroup,
@@ -910,6 +905,7 @@ int collect_tasks(void *parent, const char *controller, const char *cgroup,
 		int *alloced_pids, int *nrpids)
 {
 	char path[MAXPATHLEN];
+	nih_local char *rpath = NULL;
 
 	if (!sane_cgroup(cgroup)) {
 		nih_error("%s: unsafe cgroup", __func__);
@@ -928,7 +924,8 @@ int collect_tasks(void *parent, const char *controller, const char *cgroup,
 		return -2;
 	}
 
-	return do_collect_tasks(parent, path, pids, alloced_pids, nrpids);
+	rpath = NIH_MUST( nih_strdup(NULL, path) );
+	return do_collect_tasks(parent, &rpath, pids, alloced_pids, nrpids);
 }
 
 int get_tasks_recursive_main(void *parent, const char *controller,
@@ -1003,7 +1000,7 @@ int list_children_main(void *parent, const char *controller, const char *cgroup,
 		return -1;
 	}
 
-	return get_directory_children(parent, path, output, DT_DIR);
+	return get_directory_children(parent, path, output);
 }
 
 int do_remove_on_empty_main(const char *controller, const char *cgroup,
@@ -1228,6 +1225,42 @@ int list_controllers_main(void *parent, char ***output)
 	return 0;
 }
 
+int list_keys_main(void *parent, const char *controller, const char *cgroup,
+			struct ucred p, struct ucred r,
+			struct keys_return_type ***output)
+{
+	char path[MAXPATHLEN];
+	int nrkeys;
+
+	*output = NULL;
+	if (!sane_cgroup(cgroup)) {
+		nih_error("%s: unsafe cgroup", __func__);
+		return -1;
+	}
+
+	if (!compute_pid_cgroup(r.pid, controller, cgroup, path, NULL)) {
+		nih_error("%s: Could not determine the requested cgroup (%s:%s)",
+                __func__, controller, cgroup);
+		return -1;
+	}
+
+	/* Check access rights to the cgroup directory */
+	if (!may_access(r.pid, r.uid, r.gid, path, O_RDONLY)) {
+		nih_error("%s: Pid %d may not access %s\n", __func__, r.pid, path);
+		return -1;
+	}
+
+	nrkeys = get_directory_contents(parent, path, output);
+
+	if (nrkeys <= 0)
+		return nrkeys;
+
+	/* convert uid/gid into r's namespace */
+	convert_directory_contents(*output, r);
+
+	return nrkeys;
+}
+
 char *extra_cgroup_mounts;
 
 static int
@@ -1280,6 +1313,47 @@ static bool daemon_running(void)
 	return false;
 }
 
+static bool dir_is_sysfs(const char *path)
+{
+	struct statfs sb;
+
+	if (statfs(path, &sb) < 0)
+		// was not yet a mountpoint
+		return true;
+
+	return sb.f_type == 0x62656572;
+}
+
+static void mount_tmpfs_at(const char *path)
+{
+	if (mount("cgroup", path, "tmpfs", 0, "size=10000") < 0) {
+		nih_error("Failed to mount tmpfs at %s: %m", path);
+		exit(1);
+	}
+	nih_debug("Mounted tmpfs onto %s", CGDIR);
+}
+
+static bool is_ro_mount(const char *path)
+{
+	struct statvfs sb;
+
+	if (statvfs(path, &sb) < 0) {
+		nih_error("statvfs failed on %s: %m", path);
+		exit(1);
+	}
+
+	return sb.f_flag & MS_RDONLY;
+}
+
+static void turn_mount_rw(const char *path)
+{
+	if (mount("", path, "cgroup", MS_REMOUNT, NULL) < 0) {
+		nih_error("Failed to turn %s mount readonly: %m", path);
+		exit(1);
+	}
+	nih_debug("Turned %s from ro->rw", path);
+}
+
 /*
  * We may decide to make the socket path customizable.  For now
  * just assume it is in /sys/fs/cgroup/ which has some special
@@ -1287,36 +1361,28 @@ static bool daemon_running(void)
  */
 static bool setup_cgroup_dir(void)
 {
-	int ret;
 	if (!dir_exists(CGDIR)) {
 		nih_debug(CGDIR " does not exist");
 		return false;
 	}
+
 	if (daemon_running()) {
 		nih_error("%s: cgmanager is already running", __func__);
 		return false;
 	}
+
 	if (file_exists(CGMANAGER_SOCK)) {
 		if (unlink(CGMANAGER_SOCK) < 0) {
 			nih_error("%s: failed to delete stale cgmanager socket", __func__);
 			return false;
 		}
 	}
-	/* Check that /sys/fs/cgroup is writeable, else mount a tmpfs */
-	unlink(CGPROBE);
-	ret = creat(CGPROBE, O_RDWR);
-	if (ret >= 0) {
-		close(ret);
-		unlink(CGPROBE);
-		return mkdir_cgmanager_dir();
-	}
-	ret = mount("cgroup", CGDIR, "tmpfs", 0, "size=10000");
-	if (ret) {
-		nih_debug("Failed to mount tmpfs on %s: %s",
-			CGDIR, strerror(errno));
-		return false;
-	}
-	nih_debug("Mounted tmpfs onto %s", CGDIR);
+
+	if (dir_is_sysfs(CGDIR))
+		mount_tmpfs_at(CGDIR);
+	else if (is_ro_mount(CGDIR))
+		turn_mount_rw(CGDIR);
+
 	return mkdir_cgmanager_dir();
 }
 
