@@ -45,25 +45,41 @@
 #include <nih/main.h>
 #include <nih/logging.h>
 #include <nih/error.h>
+#include <nih/hash.h>
 
 #include <nih-dbus/dbus_connection.h>
 #include <nih-dbus/dbus_proxy.h>
 
+#include "frontend.h"  // for keys_return_type
+
 /* defines relating to the release agent */
 #define AGENT SBINDIR "/cgm-release-agent"
 #define AGENT_LINK_PATH "/run/cgmanager/agents"
 
+char *all_controllers;
+
 struct controller_mounts {
 	char *controller;
 	char *options;
 	char *path;
 	char *src;
 	char *agent;
+	struct controller_mounts *comounted;
+	bool premounted;
+	bool visited;
 };
 
 static struct controller_mounts *all_mounts;
 static int num_controllers;
 
+/*
+ * the controller_mnts is an array of the mounts to export as the controller
+ * list.  If freezer and devices are comounted, then they will form one
+ * entry "freezer,devices"
+ */
+static int num_controller_mnts;
+static char **controller_mnts;
+
 static char *base_path;
 
 bool file_exists(const char *path)
@@ -117,15 +133,10 @@ bool setup_base_run_path(void)
 
 static void set_clone_children(const char *path)
 {
-	char p[MAXPATHLEN];
+	nih_local char *p = NULL;
 	FILE *f;
-	int ret;
 
-	ret = snprintf(p, MAXPATHLEN, "%s/cgroup.clone_children", path);
-	if (ret < 0 || ret >= MAXPATHLEN) {
-		nih_fatal("%s: Error writing clone_children path", __func__);
-		return;
-	}
+	p = NIH_MUST( nih_sprintf(NULL, "%s/cgroup.clone_children", path) );
 	f = fopen(p, "w");
 	if (!f) {
 		nih_fatal("%s: Failed to set clone_children", __func__);
@@ -137,15 +148,10 @@ static void set_clone_children(const char *path)
 
 static void set_use_hierarchy(const char *path)
 {
-	char p[MAXPATHLEN];
+	nih_local char *p = NULL;
 	FILE *f;
-	int ret;
 
-	ret = snprintf(p, MAXPATHLEN, "%s/memory.use_hierarchy", path);
-	if (ret < 0 || ret >= MAXPATHLEN) {
-		nih_fatal("%s: Error writing use_hierarchy path", __func__);
-		return;
-	}
+	p = NIH_MUST( nih_sprintf(NULL, "%s/memory.use_hierarchy", path) );
 	f = fopen(p, "w");
 	if (!f) {
 		nih_fatal("%s: Failed to set memory.use_hierarchy", __func__);
@@ -155,12 +161,95 @@ static void set_use_hierarchy(const char *path)
 	fclose(f);
 }
 
+static void zero_out(struct controller_mounts *c)
+{
+	memset(c, 0, sizeof(*c));
+}
+
+/*
+ * Look for a controller_mounts struct for controller @c.
+ * If found, set @found to true and return its index in all_mounts.
+ * If not found, set @found to false and return index of the first
+ * location in all_mounts whose controller is > @c, i.e. where it
+ * should be inserted.
+ */
+static int find_controller_in_mounts(const char *c, bool *found)
+{
+	int cmp, low = 0, mid = low, high = num_controllers-1;
+
+	*found = false;
+	if (!num_controllers)
+		return 0;
+
+	while (low <= high) {
+		if (high == low)
+			break;
+		if (high == low+1) {
+			cmp = strcmp(c, all_mounts[low].controller);
+			if (cmp <= 0) {
+				mid = low;
+				break;
+			}
+			cmp = strcmp(c, all_mounts[high].controller);
+			if (cmp > 0) {
+				mid = high + 1;
+				if (mid >= num_controllers)
+					mid = num_controllers-1;
+				break;
+			}
+			mid = high;
+			break;
+		}
+		mid = low + (high-low)/2;
+		cmp = strcmp(c, all_mounts[mid].controller);
+		if (cmp == 0)
+			break;
+		if (cmp < 0)
+			high = mid;
+		else
+			low = mid;
+	}
+
+	cmp = strcmp(c, all_mounts[mid].controller);
+
+	if (cmp == 0)
+		*found = true;
+	if (cmp > 0)
+		mid++;
+
+	return mid;
+}
+
+static bool fill_in_controller(struct controller_mounts *m, char *controller,
+			char *src)
+{
+	nih_local char *dest = NULL;
+
+	dest = NIH_MUST( nih_sprintf(NULL, "%s/%s", base_path, src) );
+	m->controller = strdup(controller);
+	if (!m->controller) {
+		nih_fatal("Out of memory mounting controllers");
+		return false;
+	}
+	m->options = NULL;
+	m->path = strdup(dest);
+	m->src = strdup(src);
+	if (!m->path ||
+			!m->src) {
+		nih_fatal("Out of memory mounting controllers");
+		return false;
+	}
+	nih_info(_("Arranged to mount %s onto %s"), m->controller, m->path);
+	return true;
+}
+
 static bool save_mount_subsys(char *s)
 {
 	struct controller_mounts *tmp;
-	char *src, dest[MAXPATHLEN], *controller;
-	int ret;
+	char *src, *controller;
+	int i, insert_pt, ret;
 	size_t len = strlen(s);
+	bool found;
 
 	if (len > MAXPATHLEN) {
 		nih_fatal("bad controller type: %s", s);
@@ -168,45 +257,40 @@ static bool save_mount_subsys(char *s)
 	}
 	if ((controller = strchr(s, '='))) {
 		/* this is something like 'name=systemd' */
-		src = alloca(len+6);
+		src = alloca(len+7);
 		/* so for controller we want 'systemd' */
 		controller++;
 		/* and for source we want "none,name=systemd" */
-		snprintf(src, len+6, "none,%s", s);
+		ret = snprintf(src, len+6, "none,%s", s);
+		if (ret < 0 || ret >= len+6) {
+			nih_fatal("saving mount subsys for %s", s);
+			ret = -1;
+			goto out;
+		}
 	} else {
 		controller = s;
 		src = s;
 	}
 
-	ret = snprintf(dest, MAXPATHLEN, "%s/%s", base_path, src);
-	if (ret < 0 || ret >= MAXPATHLEN) {
-		nih_fatal("Error calculating pathname for %s and %s", base_path, src);
-		ret = -1;
-		goto out;
-	}
-	ret = -1;
+	insert_pt = find_controller_in_mounts(controller, &found);
+	if (found)
+		return true;
+
 	tmp = realloc(all_mounts, (num_controllers+1) * sizeof(*all_mounts));
 	if (!tmp) {
 		nih_fatal("Out of memory mounting controllers");
 		goto out;
 	}
 	all_mounts = tmp;
-	all_mounts[num_controllers].controller = strdup(controller);
-	if (!all_mounts[num_controllers].controller) {
-		nih_fatal("Out of memory mounting controllers");
+
+	for (i = num_controllers; i > insert_pt; i--)
+		all_mounts[i] = all_mounts[i-1];
+	zero_out(&all_mounts[insert_pt]);
+
+	if (!fill_in_controller(&all_mounts[insert_pt], controller, src)) {
+		ret = -1;
 		goto out;
 	}
-	all_mounts[num_controllers].options = NULL;
-	all_mounts[num_controllers].path = strdup(dest);
-	all_mounts[num_controllers].src = strdup(src);
-	if (!all_mounts[num_controllers].path ||
-			!all_mounts[num_controllers].src) {
-		nih_fatal("Out of memory mounting controllers");
-		goto out;
-	}
-	nih_info(_("Mounted %s onto %s"),
-			all_mounts[num_controllers].controller,
-			all_mounts[num_controllers].path);
 	num_controllers++;
 	return true;
 
@@ -217,14 +301,14 @@ out:
 static bool set_release_agent(struct controller_mounts *m)
 {
 	FILE *f;
-	char path[MAXPATHLEN];
-	int ret;
+	nih_local char *path = NULL;
 
-	ret = snprintf(path, MAXPATHLEN, "%s/release_agent", m->path);
-	if (ret < 0 || ret >= MAXPATHLEN) {
-		nih_error("out of memory");
-		return false;
+	if (m->premounted) {
+		nih_info("%s was pre-mounted, not setting a release agent",
+			m->controller);
+		return true;
 	}
+	path = NIH_MUST( nih_sprintf(NULL, "%s/release_agent", m->path) );
 	if ((f = fopen(path, "w")) == NULL) {
 		nih_error("failed to open %s for writing", path);
 		return false;
@@ -246,22 +330,35 @@ static bool set_release_agent(struct controller_mounts *m)
 static bool do_mount_subsys(int i)
 {
 	char *src, *dest, *controller;
+	struct controller_mounts *m = &all_mounts[i];
 	int ret;
 
-	dest = all_mounts[i].path;
-	controller = all_mounts[i].controller;
-	src = all_mounts[i].src;
+	dest = m->path;
+	controller = m->controller;
+	src = m->src;
 
 	if (mkdir(dest, 0755) < 0 && errno != EEXIST) {
 		nih_fatal("Failed to create %s: %s", dest, strerror(errno));
-		ret = -1;
-		goto out;
+		return false;
 	}
-	if ((ret = mount(src, dest, "cgroup", 0, src)) < 0) {
-		nih_fatal("Failed mounting %s: %s", dest, strerror(errno));
-		goto out;
+	if (m->premounted)
+		ret = mount(src, dest, "cgroup", 0, m->options);
+	else
+		ret = mount(src, dest, "cgroup", 0, src);
+	if (ret < 0) {
+		if (!m->premounted) {
+			nih_debug("Failed mounting %s onto %s: %s", src, dest, strerror(errno));
+			free(m->path);
+			m->path = NULL;
+			return true;
+		}
+		nih_fatal("Failed mounting %s onto %s: %s", src, dest, strerror(errno));
+		if (m->premounted)
+			nih_fatal("options was %s\n", m->options);
+		return false;
 	}
 	nih_info(_("Mounted %s onto %s"), controller, dest);
+
 	if (strcmp(controller, "cpuset") == 0) {
 		set_clone_children(dest); // TODO make this optional?
 		nih_info(_("set clone_children"));
@@ -270,23 +367,374 @@ static bool do_mount_subsys(int i)
 		nih_info(_("set memory.use_hierarchy"));
 	}
 
-	if (!set_release_agent(&all_mounts[i])) {
+	if (!set_release_agent(m)) {
 		nih_error("failed to set release agent for %s",
-				all_mounts[i].controller);
+				m->controller);
 		return false;
 	}
 	return true;
+}
+
+const char *controllers[] = { "blkio", "cpuset", "cpu", "cpuacct", "debug",
+			"devices", "freezer", "memory", "net_cls", "net_prio",
+			NULL };
+
+static bool is_kernel_controller(const char *c)
+{
+	int i = 0;
+
+	while (controllers[i]) {
+		if (strcmp(controllers[i++], c) == 0)
+			return true;
+	}
+	return false;
+}
+
+static bool process_mounted_subsystem(char *options)
+{
+	char *tok;
+	nih_local char *cp_opts = NULL, *cp_opts_split = NULL;
+	int i;
+	bool found;
+
+	tok = strtok(options, ",");
+	while (tok) {
+		if (strncmp(tok, "name=", 5) == 0) {
+			i = find_controller_in_mounts(tok+5, &found);
+			if (found) // jinkeys, multiple mounts already
+				goto next;
+			if (!save_mount_subsys(tok))
+				return false;
+			i = find_controller_in_mounts(tok+5, &found);
+			if (!found)
+				return false;
+			all_mounts[i].premounted = true;
+			if (!cp_opts)
+				cp_opts = NIH_MUST( nih_strdup(NULL, tok) );
+			else
+				NIH_MUST( nih_strcat_sprintf(&cp_opts, NULL, ",%s", tok) );
+		} else if (is_kernel_controller(tok)) {
+			i = find_controller_in_mounts(tok, &found);
+			if (found) // jinkeys, multiple mounts already
+				goto next;
+			if (!save_mount_subsys(tok))
+				return false;
+			i = find_controller_in_mounts(tok, &found);
+			if (!found)
+				return false;
+			all_mounts[i].premounted = true;
+			if (!cp_opts)
+				cp_opts = NIH_MUST( nih_strdup(NULL, tok) );
+			else
+				NIH_MUST( nih_strcat_sprintf(&cp_opts, NULL, ",%s", tok) );
+		}
+next:
+		tok = strtok(NULL, ",");
+	}
+
+	if (!cp_opts)
+		return true;
+
+	cp_opts_split = NIH_MUST( nih_strdup(NULL, cp_opts) );
+	tok = strtok(cp_opts_split, ",");
+	while (tok) {
+		if (strncmp(tok, "name=", 5) == 0)
+			i = find_controller_in_mounts(tok+5, &found);
+		else
+			i = find_controller_in_mounts(tok, &found);
+		if (!found)
+			return false;
+		all_mounts[i].options = strdup(cp_opts);
+		if (!all_mounts[i].options)
+			return false;
+		tok = strtok(NULL, ",");
+	}
+
+	return true;
+}
+
+/*
+ * parse /proc/self/mounts looking for already-mounted subsystems
+ */
+static bool collect_premounted_subsystems(void)
+{
+	char line[1024], *p1, *p2, *p3, *p4;
+
+	FILE *f = fopen("/proc/self/mounts", "r");
+	if (!f)
+		return false;
+	while (fgets(line, 1024, f)) {
+		p1 = strchr(line, ' ');
+		if (!p1)
+			goto bad;
+		*p1 = '\0';
+		p2 = strchr(++p1, ' ');
+		if (!p2)
+			goto bad;
+		*p2 = '\0';
+		p2++;
+		p3 = strchr(p2, ' ');
+		if (!p3)
+			goto bad;
+		*p3 = '\0';
+		if (strcmp(p2, "cgroup") != 0)
+			continue;
+		p4 = strchr(++p3, ' ');
+		if (!p4)
+			goto bad;
+		*p4 = '\0';
+		if (!process_mounted_subsystem(p3))
+			goto bad;
+	}
+	fclose(f);
+	return true;
+bad:
+	fclose(f);
+	return false;
+}
+
+static bool collate_premounted_subsystems(void)
+{
+	int i;
+
+	for (i = 0;  i < num_controllers; i++) {
+		nih_local char *opts = NULL;
+		char *tok;
+		struct controller_mounts *first = NULL, *last = NULL;
+		int j;
+		bool found;
+
+		first = &all_mounts[i];
+		if (!first->premounted)
+			continue;
+		if (first->comounted) // already linked
+			continue;
+		if (!first->options)
+			continue;
+		opts = NIH_MUST( nih_strdup(NULL, first->options) );
+		tok = strtok(opts, ",");
+		while (tok) {
+			if (strncmp(tok, "name=", 5) == 0)
+				j = find_controller_in_mounts(tok+5, &found);
+			else
+				j = find_controller_in_mounts(tok, &found);
+			if (!found)
+				return false;
+			if (!last) {
+				last = &all_mounts[j];
+				first->comounted = last;
+			} else {
+				last->comounted = &all_mounts[j];
+				last = last->comounted;
+			}
+			tok = strtok(NULL, ",");
+		}
+	}
+
+	return true;
+}
+
+static bool collect_kernel_subsystems(void)
+{
+	FILE *cgf;
+	char line[400];
+	bool bret = false;
+
+	if ((cgf = fopen("/proc/cgroups", "r")) == NULL) {
+		nih_fatal ("Error opening /proc/cgroups: %s", strerror(errno));
+		return false;
+	}
+	while (fgets(line, 400, cgf)) {
+		char *p;
+
+		if (line[0] == '#')
+			continue;
+		p = strchr(line, '\t');
+		if (!p)
+			continue;
+		*p = '\0';
+
+		if (!save_mount_subsys(line)) {
+			nih_fatal("Error storing subsystem %s", line);
+			goto out;
+		}
+	}
+	bret = true;
 
 out:
-	return false;
+	fclose(cgf);
+	return bret;
+}
+
+static void prune_from_string(char *list, char *c)
+{
+	char *f;
+	size_t len;
+	char *origlist = list;
+
+	if (strncmp(c, "none,", 5) == 0)
+		c += 5;
+	len = strlen(c);
+again:
+	if (!list)
+		return;
+	f = strstr(list, c);
+	if (!f)
+		return;
+	if (f > origlist && *(f-1) != ',') {
+		list = f+len;
+		goto again;
+	}
+	if (*(f+len) != ',' && *(f+len) != '\0') {
+		list = f+len;
+		goto again;
+	}
+	/* now we know for sure that [f-1,f+len+1] == ",f," */
+	if (f[len])
+		memmove(f, f+len+1, strlen(f+len+1)+1);
+	else
+		*(f-1) = '\0';
+	goto again;
+}
+
+static char *skip_none(char *src)
+{
+	if (strncmp(src, "none,", 5) == 0)
+		return src+5;
+	return src;
+}
+
+/*
+ * Build the list of controllers which we return as the result of
+ * ListControllers
+ */
+static void build_controller_mntlist(void)
+{
+	int i;
+	struct controller_mounts *m, *m2;
+	num_controller_mnts = 0;
+
+	for (i = 0; i < num_controllers; i++) {
+		char *srclist;
+		m = &all_mounts[i];
+		if (m->visited)
+			continue;
+		controller_mnts = realloc(controller_mnts,
+				(num_controller_mnts+1)*(sizeof(char *)));
+		if (!controller_mnts) {
+			nih_fatal("Out of memory building mntlist");
+			exit(1);
+		}
+		srclist = NIH_MUST( nih_strdup(NULL, skip_none(m->src)) );
+		m->visited = true;
+		m2 = m->comounted;
+		while (m2 && m2 != m) {
+			/*
+			 * XXX Should we use m2->controller or m2->options here?
+			 * options would include "none,".  does controller include
+			 * the 'name=' part of systemd?  If not do we need ot add it
+			 * by hande?
+			 */
+			NIH_MUST( nih_strcat_sprintf(&srclist, NULL, ",%s",
+						skip_none(m2->src)) );
+			m2->visited = true;
+			m2 = m2->comounted;
+		}
+		controller_mnts[num_controller_mnts] = srclist;
+		num_controller_mnts++;
+	}
+}
+
+static void print_debug_controller_info(void)
+{
+	int i;
+	struct controller_mounts *m;
+
+	nih_debug("all unique controllers: %s", all_controllers);
+
+	for (i = 0; i < num_controllers; i++) {
+		m = &all_mounts[i];
+		nih_debug("%d: controller %s", i, m->controller);
+		nih_debug("    src %s path %s options %s",
+			m->src, m->path ? m->path : "(none)", m->options ? m->options : "(none)");
+		nih_debug("    agent: %s", m->agent ? m->agent : "(none)");
+		nih_debug("    premounted: %s comounted: %s",
+			m->premounted ? "yes" : "no",
+			m->comounted ? m->comounted->controller : "(none)");
+	}
+}
+
+void do_list_controllers(void *parent, char ***output)
+{
+	int i;
+
+	nih_assert(output);
+	*output = NIH_MUST( nih_alloc(parent, (num_controller_mnts+1) * sizeof(char *)) );
+	(*output)[num_controller_mnts] = NULL;
+	
+	/* XXX
+	 * This will actually not be right.
+	 * if we have freezer,devices co-mounted, we'll have two separate
+	 * entries for the two.
+	 * So TODO - figure out what we want in that case, and provide it
+	 */
+	for (i = 0; i < num_controller_mnts; i++)
+		(*output)[i] = NIH_MUST( nih_strdup(parent, controller_mnts[i]) );
+}
+
+void do_prune_comounts(char *controllers)
+{
+	char *p1 = controllers, *p2;
+	int i;
+	bool found;
+	struct controller_mounts *m1;
+
+	while (p1) {
+		p2 = strchr(p1, ',');
+		if (!p2)
+			return;
+		*p2 = '\0';
+		i = find_controller_in_mounts(p1, &found);
+		if (!found) {
+			*p2 = ',';
+			p1 = p2+1;
+			continue;
+		}
+		m1 = all_mounts[i].comounted;
+		while (m1 && m1 != &all_mounts[i]) {
+			prune_from_string(p2+1, m1->src);
+			m1 = m1->comounted;
+		}
+		*p2 = ',';
+		p1 = p2+1;
+	}
+}
+
+static void build_all_controllers(void)
+{
+	int i;
+	char *c;
+
+	for (i = 0;  i < num_controllers;  i++) {
+		c = all_mounts[i].src;
+		if (strncmp(c, "none,", 5) == 0)
+			c += 5;
+		if (!all_controllers)
+			all_controllers = NIH_MUST( nih_strdup(NULL, c) );
+		else
+			NIH_MUST( nih_strcat_sprintf(&all_controllers, NULL, ",%s", c) );
+	}
+
+	do_prune_comounts(all_controllers);
 }
 
 int collect_subsystems(char *extra_mounts)
 {
-	FILE *cgf;
-	int ret;
-	char line[400];
+	/* first collect all already-mounted subsystems */
+	if (!collect_premounted_subsystems())
+		return -1;
 
+	/* handle the requested extra-mounts, which are not in /proc/cgroups */
 	if (extra_mounts) {
 		char *e;
 		for (e = strtok(extra_mounts, ","); e; e = strtok(NULL, ",")) {
@@ -297,40 +745,18 @@ int collect_subsystems(char *extra_mounts)
 		}
 	}
 
-	if ((cgf = fopen("/proc/cgroups", "r")) == NULL) {
-		nih_fatal ("Error opening /proc/cgroups: %s", strerror(errno));
+	if (!collect_kernel_subsystems())
 		return -1;
-	}
-	while (fgets(line, 400, cgf)) {
-		char *p;
-		unsigned long h;
 
-		if (line[0] == '#')
-			continue;
-		p = strchr(line, '\t');
-		if (!p)
-			continue;
-		*p = '\0';
-		h = strtoul(p+1, NULL, 10);
-		if (h) {
-			nih_info(_("%s was already mounted!"), line);
-#if STRICT
-			ret = -1;
-			goto out;
-#endif
-		}
+	if (!collate_premounted_subsystems())
+		return -1;
 
-		if (!save_mount_subsys(line)) {
-			nih_fatal("Error storing subsystem %s", line);
-			ret = -1;
-			goto out;
-		}
-	}
-	nih_info(_("found %d controllers"), num_controllers);
-	ret = 0;
-out:
-	fclose(cgf);
-	return ret;
+	build_all_controllers();
+
+	build_controller_mntlist();
+	print_debug_controller_info();
+
+	return 0;
 }
 
 /**
@@ -354,6 +780,9 @@ int setup_cgroup_mounts(void)
 		return 0;
 	}
 
+	if (mount(NULL, "/", NULL, MS_REC|MS_PRIVATE, 0) < 0)
+		nih_warn("Failed to re-mount / non-shared");
+
 	for (i=0; i<num_controllers; i++) {
 		if (!do_mount_subsys(i)) {
 			nih_fatal("Failed mounting cgroups");
@@ -390,6 +819,9 @@ bool create_agent_symlinks(void)
 	}
 
 	for (i=0; i<num_controllers; i++) {
+		if (all_mounts[i].premounted)
+			continue;
+
 		ret = snprintf(buf+plen, MAXPATHLEN-plen, "cgm-release-agent.%s",
 				all_mounts[i].controller);
 		if (ret < 0 || ret >= MAXPATHLEN) {
@@ -481,22 +913,22 @@ found:
 }
 
 /*
- * Given host @uid, return the uid to which it maps in
- * @pid's user namespace, or -1 if none.
+ * Given a open file * to /proc/pid/{u,g}id_map, and an id
+ * valid in the caller's namespace, return the id mapped into
+ * pid's namespace.
+ * Returns the mapped id, or -1 on error.
  */
-bool hostuid_to_ns(uid_t uid, pid_t pid, uid_t *answer)
+unsigned int
+convert_id_to_ns(FILE *idfile, unsigned int in_id)
 {
-	FILE *f;
-	uid_t nsuid, hostuid;
-	unsigned int count;
+	unsigned int nsuid,   // base id for a range in the idfile's namespace
+		     hostuid, // base id for a range in the caller's namespace
+		     count;   // number of ids in this range
 	char line[400];
 	int ret;
 
-	sprintf(line, "/proc/%d/uid_map", pid);
-	if ((f = fopen(line, "r")) == NULL) {
-		return false;
-	}
-	while (fgets(line, 400, f)) {
+	fseek(idfile, 0L, SEEK_SET);
+	while (fgets(line, 400, idfile)) {
 		ret = sscanf(line, "%u %u %u\n", &nsuid, &hostuid, &count);
 		if (ret != 3)
 			continue;
@@ -507,22 +939,43 @@ bool hostuid_to_ns(uid_t uid, pid_t pid, uid_t *answer)
 			 */
 			nih_error("pid wrapparound at entry %u %u %u in %s",
 				nsuid, hostuid, count, line);
-			break;
+			return -1;
 		}
-		if (hostuid <= uid && hostuid+count > uid) {
+		if (hostuid <= in_id && hostuid+count > in_id) {
 			/*
-			 * now since hostuid <= uid < hostuid+count, and
+			 * now since hostuid <= in_id < hostuid+count, and
 			 * hostuid+count and nsuid+count do not wrap around,
-			 * we know that nsuid+(uid-hostuid) which must be
+			 * we know that nsuid+(in_id-hostuid) which must be
 			 * less that nsuid+(count) must not wrap around
 			 */
-			fclose(f);
-			*answer = (uid - hostuid) + nsuid;
-			return true;
+			return (in_id - hostuid) + nsuid;
 		}
 	}
+
+	// no answer found
+	return -1;
+}
+
+/*
+ * Given host @uid, return the uid to which it maps in
+ * @pid's user namespace, or -1 if none.
+ */
+bool hostuid_to_ns(uid_t uid, pid_t pid, uid_t *answer)
+{
+	FILE *f;
+	char line[400];
+
+	sprintf(line, "/proc/%d/uid_map", pid);
+	if ((f = fopen(line, "r")) == NULL) {
+		return false;
+	}
+
+	*answer = convert_id_to_ns(f, uid);
 	fclose(f);
-	return false;
+
+	if (*answer == -1)
+		return false;
+	return true;
 }
 
 /*
@@ -592,12 +1045,18 @@ bool may_access(pid_t pid, uid_t uid, gid_t gid, const char *path, int mode)
 const char *get_controller_path(const char *controller)
 {
 	int i;
+	bool found;
 
-	for (i=0; i<num_controllers; i++) {
-		if (strcmp(all_mounts[i].controller, controller) == 0)
-			return all_mounts[i].path;
+	i = find_controller_in_mounts(controller, &found);
+	if (!found) {
+		if (strncmp(controller, "name=", 5) == 0) {
+			i = find_controller_in_mounts(controller+5, &found);
+			if (found)
+				return all_mounts[i].path;
+		}
+		return NULL;
 	}
-	return NULL;
+	return all_mounts[i].path;
 }
 
 int get_path_depth(const char *p)
@@ -641,12 +1100,16 @@ bool compute_pid_cgroup(pid_t pid, const char *controller, const char *cgroup,
 	const char *cont_path;
 	bool abspath = false;
 
-	if (!cgroup)
+	if (!cgroup) {
+		nih_error("%s: BUG: called with NULL cgroup\n", __func__);
 		return false;
+	}
 
 	if (cgroup[0] != '/') {
 		cg = pid_cgroup(pid, controller, requestor_cgpath);
 		if (!cg) {
+			nih_error("Found no cgroup entry for pid %lu controller %s\n",
+				(unsigned long)pid, controller);
 			return false;
 		}
 	} else
@@ -737,6 +1200,59 @@ out:
 	return string;
 }
 
+static int find_ordered_pid(int32_t *pids, int32_t pid, int nrpids)
+{
+	int low = 0, mid = low, high = nrpids-1;
+
+	if (!nrpids)
+		return 0;
+
+	while (low <= high) {
+		if (high == low)
+			break;
+		if (high == low+1) {
+			if (pid <= pids[low]) {
+				mid = low;
+				break;
+			}
+			if (pid > pids[high]) {
+				mid = high + 1;
+				if (mid >= nrpids)
+					mid = nrpids-1;
+				break;
+			}
+			mid = high;
+			break;
+		}
+		mid = low + (high-low)/2;
+		if (pid == pids[mid])
+			break;
+		if (pid < pids[mid])
+			high = mid;
+		else
+			low = mid;
+	}
+
+	if (pid > pids[mid])
+		mid++;
+
+	return mid;
+}
+
+static bool insert_ordered_pid(int32_t *pids, int32_t pid, int nrpids)
+{
+	int i, j;
+
+	/* TODO Switch this to binary */
+	i = find_ordered_pid(pids, pid, nrpids);
+	if (i < nrpids && pids[i] == pid)
+		return false;
+	for (j = nrpids; j > i; j--)
+		pids[j] = pids[j-1];
+	pids[i] = pid;
+	return true;
+}
+
 /*
  * file_read_pids:
  *
@@ -744,44 +1260,48 @@ out:
  * (to allow the returned value to be freed automatically when @parent is
  * freed).
  * @path: Full path to file to read.
+ * @pids: pointer to an array of ints in which pids will be placed.
+ * This should start as NULL, but on repeated calls will be re-used.
+ * @alloced_pids: size of @pids allocated so far
+ * @nrpids: number of pids in @pids which are already in use
  *
  * Read specified file and return the pids it contains.  The file is
  * expected to contain only a set of newline-separated int32_ts.
  *
- * Returns: Number of pids read, which are placed into the newly allocated
- * pids array (passed in).
+ * Returns: 0 on success, -1 on memory error, -2 on failure to open the
+ * cgroup's tasks file.
+ * @pids will be ordered
  */
-int file_read_pids(void *parent, const char *path, int32_t **pids)
+int file_read_pids(void *parent, const char *path, int32_t **pids,
+			int *alloced_pids, int *nrpids)
 {
-	int nrpids = 0, alloced_pids = 0, pid;
+	int pid;
 	FILE *fin = fopen(path, "r");
 
-	*pids = NULL;
 	if (!fin) {
 		nih_error("Error opening %s: %s", path, strerror(errno));
-		return -1;
+		return -2;
 	}
 
 	while (fscanf(fin, "%d", &pid) == 1) {
-		if (nrpids + 1 >= alloced_pids) {
+		if (*nrpids + 1 >= *alloced_pids) {
 			int32_t *tmp;
-			alloced_pids += 256;
+			*alloced_pids += 256;
 			tmp = nih_realloc(*pids, parent,
-					  alloced_pids*sizeof(int32_t));
+					  *alloced_pids*sizeof(int32_t));
 			if (!tmp) {
-				if (*pids)
-					nih_free(*pids);
-				pids = NULL;
 				nih_error("Out of memory getting pid list");
-				goto out;
+				fclose(fin);
+				return -1;
 			}
 			*pids = tmp;
+			memset(&(tmp[*nrpids]), 0, 256);
 		}
-		(*pids)[nrpids++] = (int32_t) pid;
+		if (insert_ordered_pid(*pids, (int32_t) pid, *nrpids))
+			(*nrpids)++;
 	}
-out:
 	fclose(fin);
-	return nrpids;
+	return 0;
 }
 
 /*
@@ -835,7 +1355,7 @@ void get_pid_creds(pid_t pid, uid_t *uid, gid_t *gid)
  */
 bool chown_cgroup_path(const char *path, uid_t uid, gid_t gid, bool all_children)
 {
-	int len, ret;
+	int len;
 
 	nih_assert (path);
 	len = strlen(path);
@@ -845,7 +1365,6 @@ bool chown_cgroup_path(const char *path, uid_t uid, gid_t gid, bool all_children
 	if (all_children) {
 		// chown all the files in the directory
 		struct dirent dirent, *direntp;
-		char fpath[MAXPATHLEN];
 		DIR *d;
 
 		if (len >= MAXPATHLEN)
@@ -855,13 +1374,11 @@ bool chown_cgroup_path(const char *path, uid_t uid, gid_t gid, bool all_children
 		if (!d)
 			goto out;
 
-		strcpy(fpath, path);
 		while (readdir_r(d, &dirent, &direntp) == 0 && direntp) {
+			nih_local char *fpath = NULL;
 			if (!strcmp(direntp->d_name, ".") || !strcmp(direntp->d_name, ".."))
 				continue;
-			ret = snprintf(fpath+len, MAXPATHLEN-len, "/%s", direntp->d_name);
-			if (ret < 0 || ret >= MAXPATHLEN-len)
-				continue;
+			fpath = NIH_MUST( nih_sprintf(NULL, "%s/%s", path, direntp->d_name) );
 			if (chown(fpath, uid, gid) < 0)
 				nih_error("Failed to chown file %s to %u:%u",
 					fpath, uid, gid);
@@ -1021,15 +1538,16 @@ bool realpath_escapes(char *path, char *safety)
  */
 bool move_self_to_root(void)
 {
-	char path[MAXPATHLEN];
-	int i, ret;
+	int i;
 	pid_t me = getpid();
-	FILE *f;
 
 	for (i = 0; i < num_controllers; i++) {
-		ret = snprintf(path, MAXPATHLEN, "%s/tasks", all_mounts[i].path);
-		if (ret < 0 || ret >= MAXPATHLEN)
-			return false;
+		FILE *f;
+		nih_local char *path = NULL;
+
+		if (!all_mounts[i].path)
+			continue;
+		path = NIH_MUST( nih_sprintf(NULL, "%s/tasks", all_mounts[i].path) );
 		if ((f = fopen(path, "w")) == NULL)
 			return false;
 		if (fprintf(f, "%d\n", me) <= 0) {
@@ -1043,7 +1561,7 @@ bool move_self_to_root(void)
 }
 
 /*
- * get_child_directories:
+ * get_directory_children:
  *
  * @parent: parent which will be given a reference to the returned string
  * (to allow the returned value to be freed automatically when @parent is
@@ -1056,7 +1574,7 @@ bool move_self_to_root(void)
  * Returns: Number of directories read.  The names will be placed in the
  * null-terminated array @output.
  */
-int get_child_directories(void *parent, const char *path, char ***output)
+int get_directory_children(void *parent, const char *path, char ***output)
 {
 	int used = 0, alloced = 5;
 	DIR *d;
@@ -1074,16 +1592,15 @@ int get_child_directories(void *parent, const char *path, char ***output)
 	while (readdir_r(d, &dirent, &direntp) == 0 && direntp) {
 		if (!strcmp(direntp->d_name, ".") || !strcmp(direntp->d_name, ".."))
 			continue;
-		if (DT_DIR != direntp->d_type)
+		if (direntp->d_type != DT_DIR)
 			continue;
 		if (used+1 >= alloced) {
 			char **tmp;
 			alloced += 5;
 			tmp = nih_realloc(*output, parent, alloced * sizeof(char *));
 			if (!tmp) {
-				if (*output)
-					nih_free(*output);
-				output = NULL;
+				nih_free(*output);
+				*output = NULL;
 				nih_error("%s: Out of memory", __func__);
 				closedir(d);
 				return -1;
@@ -1097,3 +1614,87 @@ int get_child_directories(void *parent, const char *path, char ***output)
 	closedir(d);
 	return used;
 }
+
+int get_directory_contents(void *parent, const char *path,
+	struct keys_return_type ***output)
+{
+	DIR *d;
+	size_t entries = 0;
+	struct dirent dirent, *direntp;
+
+	nih_assert(output);
+	d = opendir(path);
+	if (!d) {
+		nih_error("%s: failed to open directory %s: %s",
+			__func__, path, strerror(errno));
+		return -1;
+	}
+	*output = NIH_MUST( nih_alloc(parent, (entries + 1) * sizeof(**output)) );
+	(*output)[0] = NULL;
+	while (readdir_r(d, &dirent, &direntp) == 0 && direntp) {
+		struct keys_return_type *tmp;
+		struct stat sb;
+		struct keys_return_type **r;
+		nih_local char *pathname = NULL;
+
+		if (!strcmp(direntp->d_name, ".") || !strcmp(direntp->d_name, ".."))
+			continue;
+		if (direntp->d_type != DT_REG)
+			continue;
+
+		r = NIH_MUST( nih_realloc(*output, parent, (entries + 2) * sizeof(struct keys_return_type *)) );
+		*output = r;
+
+		(*output)[entries+1] = NULL;
+		(*output)[entries] = tmp = NIH_MUST( nih_new(*output, struct keys_return_type));
+		tmp->name = NIH_MUST( nih_strdup(tmp, direntp->d_name) );
+		pathname = NIH_MUST( nih_sprintf(NULL, "%s/%s", path, direntp->d_name) );
+		if (stat(pathname, &sb) < 0) {
+			tmp->uid = tmp->gid = -1;
+			tmp->perms = 0;
+		} else {
+			tmp->uid = sb.st_uid;
+			tmp->gid = sb.st_gid;
+			tmp->perms = (uint32_t) sb.st_mode;
+		}
+		entries++;
+	}
+	closedir(d);
+	return entries;
+}
+
+void convert_directory_contents(struct keys_return_type **keys, struct ucred r)
+{
+	int i = 0;
+	FILE *uidf, *gidf;
+	nih_local char *upath = NULL, *gpath = NULL;
+
+	upath = nih_sprintf(NULL, "/proc/%d/uid_map", r.pid);
+	uidf = fopen(upath, "r");
+	if (!uidf)
+		return;
+	gpath = nih_sprintf(NULL, "/proc/%d/gid_map", r.pid);
+	gidf = fopen(gpath, "r");
+	if (!gidf) {
+		fclose(uidf);
+		return;
+	}
+
+	while (keys[i]) {
+		keys[i]->uid = convert_id_to_ns(uidf, keys[i]->uid);
+		keys[i]->gid = convert_id_to_ns(gidf, keys[i]->gid);
+		i++;
+	}
+	fclose(uidf);
+	fclose(gidf);
+}
+
+
+bool was_premounted(const char *controller)
+{
+	bool found;
+	int i = find_controller_in_mounts(controller, &found);
+	if (!found)
+		return false;
+	return all_mounts[i].premounted;
+}
